import logging
import threading

import workflows
import workflows.util
import zocalo


class BaseWrapper:
    def set_recipe_wrapper(self, recwrap):
        self.recwrap = recwrap

    def prepare(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("starting", payload)

    def update(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("updates", payload)

    def done(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("completed", payload)

    def success(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("success", payload)

    def failure(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("failure", str(payload))

    def run(self):
        raise NotImplementedError()

    def record_result_individual_file(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("result-individual-file", payload)

    def record_result_all_files(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("result-all-files", payload)


class DummyWrapper(BaseWrapper):
    def run(self):
        logging.getLogger("zocalo.wrapper.DummyWrapper").info(
            "This is a dummy wrapper that simply waits for a minute."
        )
        import time

        time.sleep(10)
        self.status_thread.taskname += " (still running)"
        time.sleep(50)
        return True


class StatusNotifications(threading.Thread):
    def __init__(self, send_function, taskname):
        super().__init__(name="zocalo status notification")
        self.daemon = True
        self.send_status = send_function
        self._lock = threading.Condition(threading.Lock())
        self._status_dict = {
            "host": workflows.util.generate_unique_host_id(),
            "task": taskname,
            "workflows": workflows.version(),
            "zocalo": zocalo.__version__,
        }
        self.set_status(workflows.services.common_service.Status.STARTING)
        self._keep_running = True
        self.start()

    def set_static_status_field(self, field, value):
        """
        Add an additional static field to status notifications.
        """
        with self._lock:
            self._status_dict[field] = value

    def set_status(self, status):
        with self._lock:
            self._status_dict["status"], self._status_dict["statustext"] = (
                status.intval,
                status.description,
            )
            self._lock.notify()

    @property
    def taskname(self):
        """Return the name displayed on service monitors for this task."""
        return self._status_dict["task"]

    @taskname.setter
    def taskname(self, value):
        """Set/update the name displayed on service monitors for this task."""
        with self._lock:
            self._status_dict["task"] = value
            self._lock.notify()

    def shutdown(self):
        """Stop the status notification thread."""
        self._keep_running = False

    def run(self):
        """Status notification thread main loop."""
        with self._lock:
            self.send_status(self._status_dict)
            while self._keep_running:
                self._lock.wait(3)
                self.send_status(self._status_dict)



"""Tests related to graylog functionality."""


from unittest import mock
import zocalo


@mock.patch("zocalo.logging")
@mock.patch("zocalo.graypy")
def test_that_graypy_is_correctly_initialised(graypy, logging):
    zocalo.enable_graylog(host=mock.sentinel.host, port=mock.sentinel.port)
    logging.getLogger.return_value.addHandler.assert_called_once_with(
        graypy.GELFUDPHandler.return_value
    )
    graypy.GELFUDPHandler.assert_called_once()
    assert graypy.GELFUDPHandler.call_args[0] == (
        mock.sentinel.host,
        mock.sentinel.port,
    )


@mock.patch("zocalo.logging")
@mock.patch("zocalo.graypy")
def test_that_graypy_is_using_sensible_defaults(graypy, logging):
    zocalo.enable_graylog()
    graypy.GELFUDPHandler.assert_called_once()
    call_args = graypy.GELFUDPHandler.call_args[0]
    assert len(call_args) == 2
    assert "diamond" in call_args[0]
    assert isinstance(call_args[1], int) and call_args[1] > 0


@mock.patch("zocalo.logging")
@mock.patch("zocalo.graypy")
def test_that_python_log_levels_are_translated_to_graylog_levels(graypy, logging):
    zocalo.enable_graylog()
    assert graypy.handler.SYSLOG_LEVELS.get(10, None) == 7
    assert graypy.handler.SYSLOG_LEVELS.get(42, 20) == 3
    assert graypy.handler.SYSLOG_LEVELS.get(100, "banana") == 1



import collections
import uuid

from workflows.services.common_service import CommonService

TRACE_LOGLEVEL = 5
# a log level below debug if anyone is interested in the really-low-level spam


class Schlockmeister(CommonService):
    """
    Remove too-often-redelivered messages from the queues.

    When a message is redelivered this can have innocuous reasons, for example
    when it was processed by a service that got killed by the user, or there
    was a temporary network glitch, or anything along those lines.
    In other cases the message may cause services to die without them catching
    the error. Schlockmeister subscribes to all queues and picks out messages
    that get redelivered too often and tells the broker to quarantine them.
    Redelivered messages may not be the cause of the issue, because messages
    can be prefetched by a client, which then dies for an unrelated reason,
    causing the messages to be redelivered. For this reason the redelivery
    limit is set rather high.
    The implementation is ActiveMQ specific.

    schlockmeister, n.
    a person who deals in or sells inferior or worthless goods; junk dealer.
    """

    # Human readable service name
    _service_name = "Schlockmeister"

    # Logger name
    _logger_name = "zocalo.service.schlockmeister"

    known_queues = {}
    known_consumers = {}
    known_instances = set()

    def initializing(self):
        """
        Subscribe to all queues. Received messages must be acknowledged.
        Only receive messages that have been delivered 5 times in the past.
        Get messages without pre-processing, as this may have caused the crash.
        """

        self.log.info("Collecting broker information")

        # Generate a unique ID in order to identify myself and other instances of
        # myself in the list of active subscribers/consumers
        self.uuid = str(uuid.uuid4())

        # Listen to a specific queue in the namespace. There will be no messages
        # in that queue. It only acts as a marker to identify service instances.
        self._markerqueue = "transient.schlockmeister." + self.uuid
        self._transport.subscribe(self._markerqueue, self.ignore, disable_mangling=True)

        # Look through list of all subscriptions to identify the subscription that
        # was just set up. From this we can infer the relevant namespace.
        self._namespace = None
        self._subid_watch_global = self._transport.subscribe_broadcast(
            "ActiveMQ.Advisory.Consumer.Queue.>",
            self.watch_global,
            ignore_namespace=True,
        )

        # Add a fallback in case the namespace could not be determined.
        self._register_idle(90, self._namespace_determination_failure)

    @staticmethod
    def ignore(header, message):
        """Ignore any messages on this subscription."""

    def _namespace_determination_failure(self):
        """Report namespace determination failure."""
        self.log.error(
            "Could not determine namespace within allocated time. Shutting down."
        )
        self._shutdown()

    def watch_global(self, header, message):
        """
        Received information on one existing queue subscription.
        Try to identify the relevant transport namespace.
        """
        if self._namespace or not isinstance(message, dict):
            return
        if self._markerqueue in message.get("ConsumerInfo", {}).get(
            "destination", {}
        ).get("string", ""):
            self._namespace = message["ConsumerInfo"]["destination"]["string"][
                : -len(self._markerqueue)
            ]
            self.log.info("Identified namespace as '%s'", self._namespace)

            # With the namespace now identified, can drop the global subscription watch and look only at relevant queues
            self._transport.unsubscribe(self._subid_watch_global)

            # Disable fallback function. Enable garbage collection instead.
            self._register_idle(15, self.garbage_collect)

            self._transport.subscribe_broadcast(
                "ActiveMQ.Advisory.Consumer.Queue.%s>" % self._namespace,
                self.watch_local,
                ignore_namespace=True,
            )

    def watch_local(self, header, message):
        """Keep track of queue consumers to identify active queues and topics."""
        if not isinstance(message, dict):
            return

        if "ConsumerInfo" in message:
            destination = message["ConsumerInfo"]["destination"]["string"]
            if not destination.startswith(self._namespace):
                self.log.warning(
                    "Subscription to %s detected, which is not within namespace",
                    self._namespace,
                )
                return
            destination = destination.replace(self._namespace, "")

            subscriber = message["ConsumerInfo"]["consumerId"]["connectionId"]

            consumer_triple = (
                subscriber,
                message["ConsumerInfo"]["consumerId"]["sessionId"],
                message["ConsumerInfo"]["consumerId"]["value"],
            )
            if consumer_triple in self.known_consumers:
                self.log.error(
                    "Consumer triple %s has already been seen!", str(consumer_triple)
                )
            self.known_consumers[consumer_triple] = destination

            self.log.log(
                TRACE_LOGLEVEL, "Seen new subscriber %s to %s", subscriber, destination
            )
            if destination not in self.known_queues:
                self.known_queues[destination] = {"subscribers": collections.Counter()}
            self.known_queues[destination]["subscribers"].update({subscriber: 1})

            if destination.startswith("transient.schlockmeister."):
                self.log.info("Ignoring subscriptions by client %s", subscriber)
                self.known_instances.add(subscriber)
        elif "RemoveInfo" in message:
            subscriber = message["RemoveInfo"]["objectId"]["connectionId"]
            consumer_triple = (
                subscriber,
                message["RemoveInfo"]["objectId"]["sessionId"],
                message["RemoveInfo"]["objectId"]["value"],
            )
            if consumer_triple not in self.known_consumers:
                self.log.error(
                    "Consumer triple %s unknown for removal!", str(consumer_triple)
                )
                return
            destination = self.known_consumers[consumer_triple]
            del self.known_consumers[consumer_triple]

            if destination not in self.known_queues:
                self.log.error("Queue %s unknown for removal", destination)
            self.log.log(
                TRACE_LOGLEVEL, "Seen subscriber %s leaving %s", subscriber, destination
            )
            self.known_queues[destination]["subscribers"].update({subscriber: -1})
            if self.known_queues[destination]["subscribers"][subscriber] == 0:
                del self.known_queues[destination]["subscribers"][subscriber]
        else:
            self.log.warning("Received unknown message type\n%s", str(message))
        self.update_subscriptions()

    def update_subscriptions(self):
        """Subscribe to any new queues with real subscribers."""
        for destination in self.known_queues:
            if self.known_queues[destination].get("subscription"):
                continue
            real_subscriber_count = sum(
                map(
                    lambda k: k not in self.known_instances,
                    self.known_queues[destination]["subscribers"],
                )
            )
            if real_subscriber_count:
                self.log.debug("subscribing to %s", destination)
                self.known_queues[destination][
                    "subscription"
                ] = self._transport.subscribe(
                    destination,
                    self.quarantine,
                    acknowledgement=True,
                    selector="JMSXDeliveryCount>5",
                    disable_mangling=True,
                )

    def garbage_collect(self):
        """
        Delayed unsubscribe from lists that are without other subscribers.
        Clean up list of known queues.
        """
        queues = list(self.known_queues)
        for destination in queues:
            if self.known_queues[destination].get("subscription"):
                real_subscriber_count = sum(
                    map(
                        lambda k: k not in self.known_instances,
                        self.known_queues[destination]["subscribers"],
                    )
                )
                if not real_subscriber_count:
                    self.log.debug("unsubscribing from %s", destination)
                    self._transport.unsubscribe(
                        self.known_queues[destination]["subscription"]
                    )
                    del self.known_queues[destination]["subscription"]
            if len(self.known_queues[destination]) == 1 and not any(
                self.known_queues[destination]["subscribers"]
            ):
                del self.known_queues[destination]
                self.log.debug(
                    "collecting stale queue %s, leaving %d queues, %d consumers, %d peers",
                    destination,
                    len(self.known_queues),
                    len(self.known_consumers),
                    len(self.known_instances) - 1,
                )

    def quarantine(self, header, message):
        """Quarantine this message."""

        self.log.warning(
            "Schlockmeister has found a potentially bad message.\n"
            + "First 1000 characters of header:\n%s\n"
            + "First 1000 characters of message:\n%s",
            str(header)[:1000],
            str(message)[:1000],
        )

        # The actual quarantining magic happens on the broker
        self._transport.nack(header)



#
# zocalo.wrap
#   Wraps a command so that its status can be tracked in zocalo
#


import json
import logging
import os
import sys
from optparse import SUPPRESS_HELP, OptionParser
import pkg_resources

import workflows
import workflows.recipe.wrapper
import workflows.services.common_service
import workflows.transport
import workflows.util
import zocalo.wrapper
from zocalo import enable_graylog
from workflows.transport.stomp_transport import StompTransport


def run():
    cmdline_args = sys.argv[1:]
    # Enable logging to console
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    logging.getLogger("workflows").setLevel(logging.INFO)
    logging.getLogger("zocalo").setLevel(logging.INFO)
    logging.getLogger().setLevel(logging.WARN)
    logging.getLogger().addHandler(console)
    log = logging.getLogger("dlstbx.wrap")

    # Set up stomp defaults
    default_configuration = "/dls_sw/apps/zocalo/secrets/credentials-live.cfg"
    if "--test" in cmdline_args:
        default_configuration = "/dls_sw/apps/zocalo/secrets/credentials-testing.cfg"
    StompTransport.load_configuration_file(default_configuration)

    known_wrappers = {
        e.name: e.load for e in pkg_resources.iter_entry_points("zocalo.wrappers")
    }

    # Set up parser
    parser = OptionParser(usage="zocalo.wrap [options]")
    parser.add_option("-?", action="help", help=SUPPRESS_HELP)

    parser.add_option(
        "--wrap",
        action="store",
        dest="wrapper",
        type="choice",
        metavar="WRAP",
        default=None,
        choices=list(known_wrappers),
        help="Object to be wrapped (valid choices: %s)" % ", ".join(known_wrappers),
    )
    parser.add_option(
        "--recipewrapper",
        action="store",
        dest="recipewrapper",
        metavar="RW",
        default=None,
        help="A serialized recipe wrapper file " "for downstream communication",
    )

    parser.add_option(
        "--test", action="store_true", help="Run in ActiveMQ testing namespace (zocdev)"
    )
    parser.add_option(
        "--live",
        action="store_true",
        help="Run in ActiveMQ live namespace (zocalo, default)",
    )

    parser.add_option(
        "-t",
        "--transport",
        dest="transport",
        metavar="TRN",
        default="StompTransport",
        help="Transport mechanism. Known mechanisms: "
        + ", ".join(workflows.transport.get_known_transports())
        + " (default: %default)",
    )
    parser.add_option(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Show debug level messages",
    )
    workflows.transport.add_command_line_options(parser)

    # Parse command line arguments
    (options, args) = parser.parse_args()

    if not cmdline_args:
        parser.print_help()
        sys.exit()

    # Instantiate specific wrapper
    if not options.wrapper:
        sys.exit("A wrapper object must be specified.")

    if options.verbose:
        console.setLevel(logging.DEBUG)

    # Enable logging to graylog
    graylog_handler = enable_graylog()
    log.info(
        "Starting wrapper for %s with recipewrapper file %s",
        options.wrapper,
        options.recipewrapper,
    )

    # Connect to transport and start sending notifications
    transport = workflows.transport.lookup(options.transport)()
    transport.connect()
    st = zocalo.wrapper.StatusNotifications(transport.broadcast_status, options.wrapper)
    for env in ("SGE_CELL", "JOB_ID"):
        if env in os.environ:
            st.set_static_status_field("cluster_" + env, os.getenv(env))

    # Instantiate chosen wrapper
    instance = known_wrappers[options.wrapper]()()
    instance.status_thread = st

    # If specified, read in a serialized recipewrapper
    if options.recipewrapper:
        with open(options.recipewrapper) as fh:
            recwrap = workflows.recipe.wrapper.RecipeWrapper(
                message=json.load(fh), transport=transport
            )
        instance.set_recipe_wrapper(recwrap)

        if recwrap.environment.get("ID"):
            # If recipe ID available then include that in all future log messages
            class ContextFilter(logging.Filter):
                def filter(self, record):
                    record.recipe_ID = recwrap.environment["ID"]
                    return True

            graylog_handler.addFilter(ContextFilter())

        if recwrap.recipe_step.get("wrapper", {}).get("task_information"):
            # If the recipe contains an extra task_information field then add this to the status display
            st.taskname += (
                " (" + str(recwrap.recipe_step["wrapper"]["task_information"]) + ")"
            )

    instance.prepare("Starting processing")

    st.set_status(workflows.services.common_service.Status.PROCESSING)
    log.info("Setup complete, starting processing")

    try:
        if instance.run():
            log.info("successfully finished processing")
            instance.success("Finished processing")
        else:
            log.info("processing failed")
            instance.failure("Processing failed")
        st.set_status(workflows.services.common_service.Status.END)
    except KeyboardInterrupt:
        log.info("Shutdown via Ctrl+C")
        st.set_status(workflows.services.common_service.Status.END)
    except Exception as e:
        log.error(str(e), exc_info=True)
        instance.failure(e)
        st.set_status(workflows.services.common_service.Status.ERROR)

    instance.done("Finished processing")

    st.shutdown()
    st.join()
    log.debug("Terminating")
